"""
RT Method Module.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics import roc_auc_score
from sklearn.utils.validation import check_is_fitted


class RT(BaseEstimator):
    def __init__(
        self,
        *,
        ddof: int = 1,
        esp: float = 1e-16,
        threshold: float = 4.0,
        return_sqrt: bool = False,
    ):
        """
        RT method.

        The RT method is an unsupervised learning method used for pattern
        recognition in quality engineering. The method learns the mean of each
        feature in unit space, the sensitivity and SN ratio of each sample, and
        the associated covariance matrix of the sensitivity and SN ratio, and
        computes MD values based on these values.

        Parameters
        ----------
        ddof : int, default=1
            It means the delta degrees of freedom. The divisor used in the is
            `N - ddof`, where `N` is the number of samples.

        esp : float, default=1e-16
            A constant to avoid zero division. It is used in the calculation as
            `1 / (x + esp)`.

        threshold : float, default=4.0
            Threshold. A multiple of the standard deviation of the MD values in
            the unit space. If 4, threshold is 4 sigma.

        return_sqrt : bool, default=False
            Return the square root of the MD values or not.

        Attributes
        ----------
        mean_X_ : ndarray of shape (n_features, )
            Mean values of each feature of the training data.

        mean_Y_ : ndarray of shape (2, )
            Means of sensitivity and error variance reciprocals. Mean_Y_[0]` is
            the sensitivity mean, and Mean_Y_[1]` is the error variance
            reciprocal.

        covariance_ : ndarray of shape (2, 2)
            Variance-covariance matrix of sensitivity and error variance
            reciprocal.

        precision_ : ndarray of shape (2, 2)
            Adjoint matrix of `covariance_`.

        dist_ : ndarray of shape(n_samples, )
            Mahalanobis distances of the training set (on which the fit is
            called) observations.

        n_features_in_ : int
            Number of features seen during fit.

        feature_names_in_ : ndarray of shape (n_features_in_, )
            Names of features seen during the fit. Defined only if X has feature
            names that are all strings.
        """
        self.ddof = ddof
        self.esp = esp
        self.threshold = threshold
        self.return_sqrt = return_sqrt

    def fit(self, X, y=None):
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data.

        y : None
            Ignored.

        Returns
        -------
        self : object
            Fitted model.
        """
        X = self._validate_data(
            X=X,
            reset=True,
            ensure_min_samples=2,
            ensure_min_features=2,
            estimator=self,
        )

        self.mean_X_ = np.mean(X, axis=0)

        Y = self._compute_Y(X, self.mean_X_)

        self.mean_Y_ = np.mean(Y, axis=0)

        std_Y = Y - self.mean_Y_[None, :]

        self.covariance_ = np.cov(std_Y, rowvar=False, ddof=self.ddof)

        self.precision_ = self._get_precision(self.covariance_)

        self.dist_ = self._mahalanobis(Y, self.mean_Y_, self.precision_)

        self.sigma_ = np.sqrt(np.mean(self.dist_))

        return self

    def predict(self, X, y=None):
        """
        Predict the labels of X according to the fitted model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : None
            Ignored.

        Returns
        -------
        labels : ndarray of shape (n_samples, )
            Returns 1 for anomalies/outliers and 0 for inliers.
        """
        check_is_fitted(self)

        threshold = self.threshold * self.sigma_

        return np.where(self.mahalanobis(X=X) >= threshold, 1, 0)

    def fit_predict(self, X, y=None):
        """
        Perform Fit to X and Return Labels for X.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Input data.

        y : None
            Ignored.

        Returns
        -------
        labels : ndarray of shape (n_samples, )
            Returns 1 for anomalies/outliers and 0 for inliers.
        """
        return self.fit(X).predict(X)

    def mahalanobis(self, X):
        """
        Compute the Mahalanobis distances (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            Mahalanobis distances (MD values).
        """
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False)

        Y = self._compute_Y(X, self.mean_X_)

        MD = self._mahalanobis(Y, self.mean_Y_, self.precision_)

        if self.return_sqrt:
            MD = np.sqrt(MD)

        return MD

    def score(self, X, y):
        """
        Return the ROCAUC to the given test data and labels.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Test samples.

        y : ndarray of shape (n_samples, )
            True labels for X. 1 for anomalies/outliers and 0 for inliers.

        Returns
        -------
        score : float
            ROCAUC.
        """
        check_is_fitted(self)

        X, y = self._validate_data(X=X, y=y, reset=False)

        return roc_auc_score(y, self.mahalanobis(X=X))

    def score_samples(self, X):
        """
        Compute the Mahalanobis distances (MD values).

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            MD values.
        """
        check_is_fitted(self)

        return self.mahalanobis(X=X)

    def _compute_Y(self, X, mean_X):
        """
        Compute the sensitivity and SN ratio as Y.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        mean_X : ndarray of shape (n_features, )
            Mean values of the unit space.

        Returns
        -------
        Y : ndarray of shape (n_samples, 2)
            Sensitivity and SN Ratio. Where Y[i, 0] is the sensitivity of the
            i-th sample and Y[i, 1] is the Y[i, 1] is the SN ratio of the i-th
            sample.
        """
        std_X = X - mean_X[None, :]

        r = np.sum(mean_X**2)
        L = np.dot(std_X, mean_X)
        st = np.sum(std_X**2, axis=1)
        sb = (L**2) / (r + self.esp)
        se = st - sb
        ve = se / (std_X.shape[1] - 1)
        y1 = L / (r + self.esp)
        y2 = np.sqrt(ve)
        Y = np.hstack([y1[:, None], y2[:, None]])

        return Y

    def _get_precision(self, covariance):
        """
        Compute precision from covariance.

        Parameters
        ----------
        covariance : ndarray of shape (n_features, n_features)
            Covariance matorix of the unit space.

        Returns
        -------
        precision : ndarray of shape (n_features, n_features)
            Adjoint matrix of the covariance matrix.
        """
        precision = np.empty((2, 2))
        precision[0, 0] = covariance[1, 1]
        precision[0, 1] = -covariance[1, 0]
        precision[1, 0] = -covariance[0, 1]
        precision[1, 1] = covariance[0, 0]

        return precision

    def _mahalanobis(self, Y, mean_Y, precision):
        """
        Compute Mahalanobis distances (MD values).

        Parameters
        ----------
        Y : ndarray of shape (n_samples, 2)
            The sensitivity and the SN ratio of the samples.

        mean_Y : ndarray of shape (2, )
            Mean values of Y in the unit space.

        precision : ndarray of shape (2, 2), default=None
            Adjoint matrix of the covariance matrix in unti space.

        Returns
        -------
        MD : ndarray of shape (n_samples, )
            Mahalanobis distances (MD values).
        """
        std_Y = Y - mean_Y[None, :]

        MD = (std_Y.dot(precision) * std_Y).mean(axis=1)

        if self.return_sqrt:
            MD = np.sqrt(MD)

        return MD

    def _more_tags(self):
        return {"binary_only": True}



"""
The 'mts' module contains various methods of the MT system.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

from ._msr import MSR
from ._mt import MT
from ._rt import RT
from ._t import T


__all__ = ["MT", "RT", "T", "MSR"]



"""
MSR Method Module.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.model_selection import KFold
from sklearn.utils.validation import check_is_fitted


class MSR(RegressorMixin, BaseEstimator):
    def __init__(self, *, delta: float = 1e-4, esp: float = 1e-16):
        """
        MSR: Multiple Single Regression.

        Parameters
        ----------
        delta : float, default=1e-4
            Threshold for stopping repeated computations.

        esp : float, default=1e-16
            A constant to avoid zero division. It is used in the calculation as
            `1 / (x + esp)`.

        Attributes
        ----------
        mean_X_ : ndarray of shape(n_features, )
            Mean values of each feature of the training data.

        mean_y_ : float
            Mean value of target values.

        coef_ : ndarray of shape (n_features, )
            Estimated coefficients for the MSR.

        n_features_in_ : int
            Number of features seen during fit.

        feature_names_in_ : ndarray of shape (n_features_in_, )
            Names of features seen during the fit. Defined only if X has feature
            names that are all strings.

        References
        ----------
        前田誠. (2017). T 法 (1) の考え方を利用した新しい回帰手法の提案. 品質, 47(2),
        185-194.
        """
        self.delta = delta
        self.esp = esp

    def _compute_sn_ratio_and_sensitivity(self, X, y):
        r = np.sum(X**2, axis=0)
        L = np.dot(X.T, y)
        st = np.dot(y, y)
        sb = (L**2) / (r + self.esp)
        se = st - sb
        n = np.sqrt(sb / (se + self.esp))
        b = L / (r + self.esp)

        return st, sb, n, b

    def fit(self, X, y):
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data.

        y : ndarray of shape (n_samples, )
            Target values. Will be cast to X's dtype if necessary.

        Returns
        -------
        self : object
            Fitted model.
        """
        X, y = self._validate_data(
            X=X,
            y=y,
            reset=True,
            y_numeric=True,
            ensure_min_samples=2,
            estimator=self,
        )

        n_samples, n_features = X.shape

        if n_samples <= 50:
            n_splits = n_samples
        else:
            n_splits = int(2250 / n_samples) + 5

        kf = KFold(n_splits=n_splits)

        self.coef_ = np.zeros(n_features)
        coef_kf = np.zeros((n_splits, n_features))

        self.mean_X_ = np.mean(X, axis=0)
        self.mean_y_ = np.mean(y)

        std_X = X - self.mean_X_[None, :]
        std_y = y - self.mean_y_

        zz_before = None
        skip_kf = []
        while True:
            y_ = np.dot(std_X, self.coef_)

            z = std_y - y_

            st, sb, n, b = self._compute_sn_ratio_and_sensitivity(std_X, z)

            if st == 0 or np.all(sb == 0):
                break

            self.coef_ += b * n / np.sum(n)

            z = np.empty(n_samples)
            for kf_idx, (train_idx, test_idx) in enumerate(kf.split(std_X)):
                train_X, train_y = std_X[train_idx], std_y[train_idx]
                test_X, test_y = std_X[test_idx], std_y[test_idx]

                if kf_idx in skip_kf:
                    y_kf = np.dot(test_X, coef_kf[kf_idx])
                    z[test_idx] = test_y - y_kf
                    continue

                y_kf = np.dot(train_X, coef_kf[kf_idx])
                z_kf = train_y - y_kf

                st, sb, n, b = self._compute_sn_ratio_and_sensitivity(train_X, z_kf)

                if st == 0 or np.all(sb == 0):
                    skip_kf.append(kf_idx)
                    y_kf = np.dot(test_X, coef_kf[kf_idx])
                    z[test_idx] = test_y - y_kf
                    continue

                coef_kf[kf_idx] += b * n / np.sum(n)

                y_kf = np.dot(test_X, coef_kf[kf_idx])
                z[test_idx] = test_y - y_kf

            zz_after = np.dot(z, z)

            if zz_before is None:
                zz_before = zz_after * 2

            if (zz_before - zz_after) <= (self.delta * zz_before):
                break
            else:
                zz_before = zz_after

        return self

    def predict(self, X, y=None):
        """
        Predict using the fitted model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : None
            Ignored.

        Returns
        -------
        y_pred : ndarray of shape (n_samples, )
            Predicted values.
        """
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False)

        std_X = X - self.mean_X_[None, :]

        return np.dot(std_X, self.coef_) + self.mean_y_

    def _more_tags(self):
        return RegressorMixin._more_tags(self)



"""
T(1), T(2), Ta and Tb Methods Module.
"""

# Authors: Shota Fukuda <st_fukuda@outlook.jp>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted


class T(RegressorMixin, BaseEstimator):
    def __init__(
        self, *, tb: bool = False, esp: float = 1e-16, is_simplified: bool = False
    ):
        """
        T(1), T(2), Ta and Tb methods.

        The T(1), T(2), Ta and Tb methods are supervised learning methods used
        for regression in quality engineering. The T(1) and T(2) methods divide
        the training data into unit space and signal data, and learn the mean
        from the unit space and the sensitivity and SN ratio from the signal
        data. The Ta method does not divide the training data into unit space
        and signal data, and learns the mean, sensitivity, and SN ratio from all
        the training data. The Tb method also learns from all training data, but
        for each element, the sample with the largest SN ratio is used as the
        mean.

        Parameters
        ----------
        tb : bool, default=False
            Whether to compute as Tb method. If False, compute as T(1), T(2),
            and Ta methods.

        esp : float, default=1e-16
            A constant to avoid zero division. It is used in the calculation as
            `1 / (x + esp)`.

        is_simplified : bool, default=False
            Compute the SN ratio using the simplified formula or not. The
            simplified formula computes with `b**2 / ve`.

        Attributes
        ----------
        mean_X_ : ndarray of shape (n_features, )
            Mean values of each feature of the training data.

        mean_y_ : float or ndarray of shape (n_features, )
            Mean value of target values.

        n_ : ndarray of shape (n_features, )
            SN ratio between each feature and the target values.

        b_ : ndarray of shape (n_features, )
            Sensitivity between each feature and target values.

        n_features_in_ : int
            Number of features seen during fit.

        feature_names_in_ : ndarray of shape (n_features_in_, )
            Names of features seen during the fit. Defined only if X has feature
            names that are all strings.
        """
        self.tb = tb
        self.esp = esp
        self.is_simplified = is_simplified

    def fit(self, X, y, *, us_idx=None):
        """
        Fit the model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Training data. Includes unit space and signal data.

        y : ndarray of shape (n_samples, )
            Target values. Will be cast to X's dtype if necessary.

        us_idx : array_like of shape (n_samples, ) or None, default=None.
            A binary array indicating which sample of the training data is the
            unit space (0 for the unit space, 1 for the signal data); if None,
            the training data is not divided into the unit space and the signal
            data, but is computed as the Ta method. It is ignored when the Tb
            method is computed.

        Returns
        -------
        self : object
            Fitted model.
        """
        X, y = self._validate_data(
            X=X,
            y=y,
            reset=True,
            y_numeric=True,
            estimator=self,
        )

        if self.tb:
            n = np.empty_like(X)
            b = np.empty_like(X)
            for i, (x_i, y_i) in enumerate(zip(X, y)):
                std_X = X - x_i[None, :]
                std_y = y - y_i

                n[i], b[i] = self._compute_sn_ratio_and_sensitivity(std_X, std_y)

            idx_row = np.argmax(n, axis=0)
            idx_col = np.arange(X.shape[1])

            self.mean_X_ = X[idx_row, idx_col]
            self.mean_y_ = y[idx_row]

            self.b_ = b[idx_row, idx_col]
            self.n_ = n[idx_row, idx_col]
        else:
            if us_idx is None:
                self.mean_X_ = np.mean(X, axis=0)
                self.mean_y_ = np.mean(y)

                std_X = X - self.mean_X_[None, :]
                std_y = y - self.mean_y_
            else:
                unit_space_mask = np.where(us_idx == 0, True, False)

                self.mean_X_ = np.mean(X[unit_space_mask], axis=0)
                self.mean_y_ = np.mean(y[unit_space_mask])

                std_X = X[~unit_space_mask] - self.mean_X_[None, :]
                std_y = X[~unit_space_mask] - self.mean_y_

            self.n_, self.b_ = self._compute_sn_ratio_and_sensitivity(std_X, std_y)

        return self

    def predict(self, X, y=None):
        """
        Predict using the fitted model.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : None
            Ignored.

        Returns
        -------
        y_pred : ndarray of shape (n_samples, )
            Predict values.
        """
        check_is_fitted(self)

        X = self._validate_data(X=X, reset=False, estimator=self)

        std_X = X - self.mean_X_

        M_pred = std_X / (self.b_ + self.esp)[None, :]

        if self.tb:
            y_pred = M_pred + self.mean_y_[None, :]  # type: ignore
            y_pred = np.dot(y_pred, self.n_) / (np.sum(self.n_) + self.esp)
        else:
            M_pred = np.dot(M_pred, self.n_) / (np.sum(self.n_) + self.esp)
            y_pred = M_pred + self.mean_y_

        return y_pred

    def score(self, X, y):
        """
        Return the SN ratio of the integrated estimate.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Test samples.

        y : ndarray of shape (n_samples, )
            True values for X.

        Returns
        -------
        n : float
            SN ratio of the integrated estimate. It is computed from M_True and
            M_Pred for the T(1), T(2) and Ta methods, and from y_True and y_Pred
            for the Tb method.
        """
        check_is_fitted(self)

        X, y = self._validate_data(X=X, y=y, reset=False)

        if self.tb:
            M_true = y
            M_pred = self.predict(X)
        else:
            M_true = y - self.mean_y_
            M_pred = self.predict(X) - self.mean_y_

        n, _ = self._compute_sn_ratio_and_sensitivity(M_pred[:, None], M_true)
        n = 10 * np.log10(n)

        return n

    def _compute_sn_ratio_and_sensitivity(self, X, y):
        """
        Compute the SN ratio and sensitivity.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Samples.

        y : ndarray of shape (n_samples, )
            Target values.

        Returns
        -------
        n : ndarray of shape (n_features, )
            SN ratio between each feature and the target values.

        b : ndarray of shape (n_features, )
            Sensitivity between each feature and the target values.
        """
        r = np.sum(y**2)
        L = np.dot(X.T, y)
        st = np.sum(X**2, axis=0)
        sb = (L**2) / (r + self.esp)
        se = st - sb
        ve = se / (X.shape[0] - 1)

        b = L / (r + self.esp)

        if self.is_simplified:
            n = (b**2) / (ve + self.esp)
        else:
            n = (1 / r * (sb - ve)) / (ve + self.esp)
            mask = sb <= ve
            n[mask] = 0

        return n, b

    def _more_tags(self):
        return RegressorMixin._more_tags(self)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mts']

package_data = \
{'': ['*']}

install_requires = \
['scikit-learn']

setup_kwargs = {
    'name': 'mt-system',
    'version': '0.1.9',
    'description': 'Python library of MT system.',
    'long_description': '# MT-system\n\nPython library of MT system.\n\n## Dependencies\n\nThis library requires:\n\n- Python (>=3.8)\n- scikit-learn (>=1.1.0)\n\n## Installation\n\nIt can be installed as follows using pip:\n\n```shell\npip install -U mt-system\n```\n\n## Usage\n\nDescribe how to use the library in this part.\n\n## Development\n\nDevelopment requires:\n\n- Python (>=3.8)\n- Poetry\n- Git\n- Make (Option)\n\n### Source code\n\nYou can check the latest sources with the command:\n\n```shell\ngit clone https://github.com/stfukuda/mt-system.git\n```\n\n### Enviroment\n\nAfter cloning the repository, you can install the development environment with the command:\n\n```shell\nmake install\n```\n\n### Testing\n\nAfter installation, you can run the test with the command:\n\n```shell\nmake test\n```\n\n### Submitting a Pull Request\n\nIf the test passes, send the pull request according to the format.\n\n## License\n\n[BSD-3-Clause License](LICENSE)\n',
    'author': 'Shota Fukuda',
    'author_email': 'st_fukuda@outlook.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/stfukuda/mt-system.git',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
